#ifndef _RHEO_GRAD_H
#define _RHEO_GRAD_H
///
/// This file is part of Rheolef.
///
/// Copyright (C) 2000-2009 Pierre Saramito <Pierre.Saramito@imag.fr>
///
/// Rheolef is free software; you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation; either version 2 of the License, or
/// (at your option) any later version.
///
/// Rheolef is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with Rheolef; if not, write to the Free Software
/// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
/// 
/// =========================================================================

/*Class:grad
NAME: @code{grad} -- gradient operator
@bfindex grad
@cindex gradient
@apindex P1
@apindex P2
@apindex P1d
@apindex P0
SYNOPSIS:
    @example
      form(const space V, const space& M, "grad");
    @end example
DESCRIPTION:       
    @noindent
    Assembly the form associated to the
    gradient operator on a finite element space @code{V}:
@iftex  
@tex
    $$ 
      b(u, {\bf q}) = \int_\Omega \nabla u . {\bf q} \, dx
    $$ 
@end tex
@end iftex  
@ifnottex  
    @example
                /
                |
      b(u, q) = | grad(u).q dx
                |
                / Omega
    @end example
@end ifnottex  
    @noindent
    The V space may be a either
    @code{P1}  or @code{P2}  finite element space,
    while the M space may be @code{P0}  or @code{P1d}  respectively.
    See also @ref{form class} and @ref{space class}.

EXAMPLE:
    @noindent
    The following piece of code build the divergence form
    associated to the @code{P1} approximation:
    @example
   	geo omega("square");
   	space V(omega, "P1");
   	space M(omega, "P0", "vector");
   	form b(V, M, "grad");
    @end example
AUTHOR: 
    LMC-IMAG, 38041 Grenoble cedex 9, France
    | Pierre.Saramito@imag.fr
DATE:
    3 march 2001
End:
*/
#include "rheolef/form_element_rep.h"
namespace rheolef { 

class grad: public form_element_rep {

public:
// allocator:

    grad() : form_element_rep() {}

// virtual accessor:

    void operator() (const geo_element& K, ublas::matrix<Float>& m) const;
    size_type n_derivative() const;
    void check_after_initialize () const;
};
}// namespace rheolef
#endif // _RHEO_GRAD_H
